
/*******************************************************************************

 Copyright (c) 2009, Charles McGarvey
 All rights reserved.
 
 Redistribution   and   use  in  source  and  binary  forms,  with  or  without
 modification, are permitted provided that the following conditions are met:
 
   * Redistributions  of  source  code  must retain the above copyright notice,
     this list of conditions and the following disclaimer.
   * Redistributions  in binary form must reproduce the above copyright notice,
     this  list of conditions and the following disclaimer in the documentation
     and/or other materials provided with the distribution.
 
 THIS  SOFTWARE  IS  PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 AND  ANY  EXPRESS  OR  IMPLIED  WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 DISCLAIMED.  IN  NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
 FOR  ANY  DIRECT,  INDIRECT,  INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 DAMAGES  (INCLUDING,  BUT  NOT  LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 SERVICES;  LOSS  OF  USE,  DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
 CAUSED  AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 OR  TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

*******************************************************************************/

#include <SDL/SDL_image.h>

#include "Dispatcher.hh"
#include "Exception.hh"
#include "Log.hh"
#include "Settings.hh"
#include "Video.hh"


namespace Mf {


Video::Video()
{
	init(mAttribs);
}

Video::Video(const Attributes& attribs)
{
	init(attribs);
}

Video::Video(const std::string& caption, const std::string& icon)
{
	if (mAttribs.caption == "Untitled")
	{
		mAttribs.caption = caption;
	}
	if (mAttribs.icon == "")
	{
		mAttribs.icon = icon;
	}

	init(mAttribs);
}

void Video::init(const Attributes& attribs)
{
	mContext = 0;
	mFlags = 0;
	mAttribs = attribs;

	setFull(attribs.fullscreen);
	setResizable(attribs.resizable);
	setOpenGLAttributes();
	setCaption(attribs.caption);
	setIcon();
	setCursorVisible(attribs.cursorVisible);
	setCursorGrab(attribs.cursorGrab);
	setVideoMode(attribs.mode);
}

void Video::recreateContext()
{
	SDL_FreeSurface(mContext);
	mContext = 0;
	setVideoMode(mAttribs.mode);
}

void Video::setOpenGLAttributes()
{
	SDL_GL_SetAttribute(SDL_GL_RED_SIZE,           mAttribs.colorBuffer[0]);
	SDL_GL_SetAttribute(SDL_GL_GREEN_SIZE,         mAttribs.colorBuffer[1]);
	SDL_GL_SetAttribute(SDL_GL_BLUE_SIZE,          mAttribs.colorBuffer[2]);
	SDL_GL_SetAttribute(SDL_GL_ALPHA_SIZE,         mAttribs.colorBuffer[3]);
	SDL_GL_SetAttribute(SDL_GL_BUFFER_SIZE,        mAttribs.frameBuffer);
	SDL_GL_SetAttribute(SDL_GL_DOUBLEBUFFER,       mAttribs.doubleBuffer);
	SDL_GL_SetAttribute(SDL_GL_DEPTH_SIZE,         mAttribs.depthBuffer);
	SDL_GL_SetAttribute(SDL_GL_STENCIL_SIZE,       mAttribs.stencilBuffer);
	SDL_GL_SetAttribute(SDL_GL_ACCUM_RED_SIZE,     mAttribs.accumBuffer[0]);
	SDL_GL_SetAttribute(SDL_GL_ACCUM_GREEN_SIZE,   mAttribs.accumBuffer[1]);
	SDL_GL_SetAttribute(SDL_GL_ACCUM_BLUE_SIZE,    mAttribs.accumBuffer[2]);
	SDL_GL_SetAttribute(SDL_GL_ACCUM_ALPHA_SIZE,   mAttribs.accumBuffer[3]);
	SDL_GL_SetAttribute(SDL_GL_STEREO,             mAttribs.stereo);
	SDL_GL_SetAttribute(SDL_GL_MULTISAMPLEBUFFERS, mAttribs.multisampleBuffers);
	SDL_GL_SetAttribute(SDL_GL_MULTISAMPLESAMPLES, mAttribs.multisampleSamples);
	SDL_GL_SetAttribute(SDL_GL_SWAP_CONTROL,       mAttribs.swapControl);
	SDL_GL_SetAttribute(SDL_GL_ACCELERATED_VISUAL, mAttribs.hardwareonly);
}


Video::~Video()
{
	SDL_FreeSurface(mContext);
}


void Video::setVideoMode(const long mode[3])
{
	if (mode != mAttribs.mode || !mContext)
	{
		if (mContext) SDL_FreeSurface(mContext);

		mContext = SDL_SetVideoMode(mode[0], mode[1], mode[2],
				SDL_OPENGL | mFlags);

		if (mContext)
		{
			mAttribs.mode[0] = mode[0];
			mAttribs.mode[1] = mode[1];
			mAttribs.mode[2] = mode[2];

#if defined(_WIN32) || defined(__WIN32__)
			// on win32, creating a new context via SDL_SetVideoMode will wipe
			// out the GL state, so we gotta notify everyone to reload their
			// state after the change
			Mf::dispatcher::dispatch("video.context_recreated");
			logInfo("video context recreated");
#endif
		}
		else throw Exception(ErrorCode::SDL_VIDEOMODE);
	}
}

Video::Attributes Video::getAttributes() const
{
	return mAttribs;
}


void Video::resize(int width, int height)
{
	long mode[] = {width, height, mAttribs.mode[2]};
	setVideoMode(mode);
}

bool Video::iconify()
{
	return SDL_WM_IconifyWindow();
}


void Video::setCaption(const std::string& caption)
{
	mAttribs.caption = caption;
	SDL_WM_SetCaption(caption.c_str(), 0);
}

void Video::setIcon()
{
	if (mAttribs.icon != "")
	{
		SDL_Surface* icon = IMG_Load(mAttribs.icon.c_str());
		if (icon)
		{
			SDL_WM_SetIcon(icon, 0);
			SDL_FreeSurface(icon);
		}
	}
}

std::string Video::getCaption() const
{
	return mAttribs.caption;
}


void Video::setFull(bool full)
{
	if (full != isFull() || !mContext)
	{
		if (mContext)
		{
			mFlags ^= SDL_FULLSCREEN;

#if defined(linux) || defined(__linux) || defined(__linux__)
			if (SDL_WM_ToggleFullScreen(mContext) == 0)
#endif
			recreateContext();
		}
		else
		{
			if (full) mFlags |= SDL_FULLSCREEN;
			else mFlags &= ~SDL_FULLSCREEN;
		}
	}
}

void Video::toggleFull()
{
	setFull(!isFull());
}

bool Video::isFull() const
{
	return mFlags & SDL_FULLSCREEN;
}


void Video::setCursorVisible(bool hasCursor)
{
	SDL_ShowCursor(hasCursor? SDL_ENABLE : SDL_DISABLE);
}

void Video::toggleCursorVisible()
{
	setCursorVisible(!isCursorVisible());
}

bool Video::isCursorVisible() const
{
	return (SDL_ShowCursor(SDL_QUERY) == SDL_ENABLE);
}


void Video::setResizable(bool resizable)
{
	if (resizable != isResizable() || !mContext)
	{
		if (mContext)
		{
			mFlags ^= SDL_RESIZABLE;
			recreateContext();
		}
		else
		{
			if (resizable) mFlags |= SDL_RESIZABLE;
			else mFlags &= ~SDL_RESIZABLE;
		}
	}
}

void Video::toggleResizable()
{
	setResizable(!isResizable());
}

bool Video::isResizable() const
{
	return mFlags & SDL_RESIZABLE;
}


bool Video::isCursorGrab() const
{
	return (SDL_WM_GrabInput(SDL_GRAB_QUERY) == SDL_GRAB_ON);
}

void Video::toggleCursorGrab()
{
	setCursorGrab(!isCursorGrab());
}

void Video::setCursorGrab(bool cursorGrab)
{
	SDL_WM_GrabInput(cursorGrab? SDL_GRAB_ON : SDL_GRAB_OFF);
}


void Video::makeActive()
{
	// NOP until the day SDL supports more than only one window.
	// Still waiting...
}

void Video::swap()
{
	SDL_GL_SwapBuffers();
}


int Video::getWidth() const
{
	return mContext->w;
}

int Video::getHeight() const
{
	return mContext->h;
}


Video::Attributes::Attributes()
{
	// Set some sane GL and window defaults (see SDL_video.c:217)
	colorBuffer[0] = 3;
	colorBuffer[1] = 3;
	colorBuffer[2] = 2;
	colorBuffer[3] = 0;
	frameBuffer = 0;
	doubleBuffer = true;
	depthBuffer = 16;
	stencilBuffer = 0;
	accumBuffer[0] = 0;
	accumBuffer[1] = 0;
	accumBuffer[2] = 0;
	accumBuffer[3] = 0;
	stereo = false;
	multisampleBuffers = 0;
	multisampleSamples = 0;
	swapControl = false;
	hardwareonly = false;
	mode[0] = 640;
	mode[1] = 480;
	mode[2] = 0;
	fullscreen = false;
	resizable = false;
	cursorVisible = true;
	cursorGrab = false;

	Settings& settings = Settings::getInstance();

	std::vector<long> colors;
	settings.get("colorbuffers", colors);
	if (colors.size() > 0) colorBuffer[0] = colors[0];
	if (colors.size() > 1) colorBuffer[1] = colors[1];
	if (colors.size() > 2) colorBuffer[2] = colors[2];
	if (colors.size() > 3) colorBuffer[3] = colors[3];

	settings.get("framebuffer", frameBuffer);
	settings.get("doublebuffer", doubleBuffer);
	settings.get("depthbuffer", depthBuffer);
	settings.get("stencilbuffer", stencilBuffer);

	std::vector<long> accum;
	settings.get("accumbuffers", accum);
	if (accum.size() > 0) accumBuffer[0] = accum[0];
	if (accum.size() > 1) accumBuffer[1] = accum[1];
	if (accum.size() > 2) accumBuffer[2] = accum[2];
	if (accum.size() > 3) accumBuffer[3] = accum[3];

	settings.get("stereo", stereo);
	settings.get("multiesamplebuffers", multisampleBuffers);
	settings.get("multiesamplesamples", multisampleSamples);
	settings.get("swapcontrol", swapControl);
	settings.get("hardwareonly", hardwareonly);

	if (!settings.get("caption", caption))
	{
		caption = "Untitled";
	}
	settings.get("icon", icon);

	std::vector<long> dimensions;
	settings.get("videomode", dimensions);
	if (dimensions.size() > 1)
	{
		mode[0] = dimensions[0];
		mode[1] = dimensions[1];
	}
	if (dimensions.size() > 2) mode[2] = dimensions[2];

	settings.get("fullscreen", fullscreen);
	settings.get("resizable", resizable);
	settings.get("showcursor", cursorVisible);
	settings.get("grab", cursorGrab);
}


} // namespace Mf

/** vim: set ts=4 sw=4 tw=80: *************************************************/

